@@ -5,10 +5,9 @@ Gestures wrapper for iOS
 
 # Gestures for the Pythonista iOS app
  
-This is a convenience class for enabling gestures in Pythonista UI
-applications, including built-in views. Main intent here has been to make
-them Python friendly, hiding all the Objective-C stuff. All gestures
-correspond to the standard Apple gestures.
+This is a convenience class for enabling gestures, including drag and drop
+support, in Pythonista UI applications. Main intent here has been to make
+them Python friendly, hiding all the Objective-C details.
 
 Run the file on its own to see a demo of the supported gestures.
 
@@ -22,6 +21,14 @@ Copy from [GitHub](https://github.com/mikaelho/pythonista-gestures), or
 
 with [stash](https://github.com/ywangd/stash).
 
+## Versions:
+
+* 1.2 - Adds drag and drop support.  
+* 1.1 - Adds distance parameters to swipe gestures.
+* 1.0 - First version released to PyPi. 
+  Breaks backwards compatibility in syntax, adds multi-recognizer coordination,
+  and removes force press support.
+
 ## Usage
 
 For example, do something when user swipes left on a Label:
@@ -56,7 +63,36 @@ All of the gesture-adding methods return an object that can be used
 to remove or disable the gesture as needed, see the API. You can also remove
 all gestures from a view with `remove_all_gestures(view)`.
 
-#docgen-toc
+## Drag and drop
+
+This module supports dragging and dropping both within a Pythonista app and
+between Pythonista and another app (only possible on iPads). These two cases
+are handled differently:
+    
+* For in-app drops, Apple method of relaying objects is skipped completely,
+  and you can refer to _any_ Python object to be dropped to the target view.
+* For cross-app drops, we have to conform to Apple method of managing data.
+  Currently only plain text and image drops are supported, in either direction.
+* It is also good to note that `ui.TextField` and `ui.TextView` views natively
+  act as receivers for both in-app and cross-app plain text drag and drop.
+
+View is set to be a sender for a drap and drop operation with the `drag`
+function. Drag starts with a long press, and can end in any view that has been
+set as a receiver with the `drop` function. Views show the readiness to receive
+data with a green "plus" sign. You can accept only specific types of data;
+incompatible drop targets show a grey "forbidden" sign.
+
+Following example covers setting up an in-app drag and drop operation between
+two labels. To repeat, in the in-app case, the simple string could replaced by
+any Python object of any complexity, passed by reference:
+    
+    drag(sender_label, "Important data")
+    
+    drop(receiver_label,
+        lambda data, sender, receiver: setattr(receiver, 'text', data),
+        accept=str)
+
+See the documentation for the two functions for details.
 
 ## Fine-tuning gesture recognition
 
@@ -119,18 +155,15 @@ phone use:
   recognizer to e.g. ui.Label produces no results.
 * It can be hard to add gestures to ui.ScrollView, ui.TextView and the like,
   because they have complex multi-view structures and gestures already in
-  place.
-  
-## Versions:
-    
-* 1.1 - Adds distance parameters to swipe gestures.
-* 1.0 - First version released to PyPi. 
-  Breaks backwards compatibility in syntax, adds multi-recognizer coordination,
-  and removes force press support.
+  place.  
 """
 
-__version__ = '1.1'
+__version__ = '1.2'
 
+import ctypes
+import functools
+import inspect
+import os
 import types
 
 import ui
@@ -146,6 +179,18 @@ UIPinchGestureRecognizer = ObjCClass('UIPinchGestureRecognizer')
 UIRotationGestureRecognizer = ObjCClass('UIRotationGestureRecognizer')
 UISwipeGestureRecognizer = ObjCClass('UISwipeGestureRecognizer')
 
+#  Drag and drop classes
+
+NSItemProvider = ObjCClass('NSItemProvider')
+UIDragItem = ObjCClass('UIDragItem')
+UIDragInteraction = ObjCClass('UIDragInteraction')
+UIDropInteraction = ObjCClass('UIDropInteraction')
+UIDropProposal = ObjCClass('UIDropProposal')
+NSItemProvider = ObjCClass('NSItemProvider')
+UIImagePNGRepresentation = c.UIImagePNGRepresentation
+UIImagePNGRepresentation.restype = c_void_p
+UIImagePNGRepresentation.argtypes = [c_void_p]
+
 # Constants
 
 # Recognizer states
@@ -174,11 +219,13 @@ EDGE_BOTTOM = 4
 EDGE_RIGHT = 8
 EDGE_ALL = 15
 
+
 class Data():
     """
     Simple class that contains all the data about the gesture. See the Usage
     section and individual gestures for information on the data included. 
     Also provides convenience state-specific properties (`began` etc.).
+    (docgen-ignore)
     """
     
     def __init__(self):
@@ -229,70 +276,122 @@ class Data():
     def failed(self):
         return self.state == FAILED
             
-def is_objc_type(objc_instance, objc_class):
-    return objc_instance.isKindOfClass_(objc_class.ptr)
             
-def gestureAction(_self, _cmd):
-    slf = ObjCInstance(_self)
-    view = slf.view
-    recognizer = slf.recognizer
-    handler_func = slf.handler_func
-    data = Data()
-    data.recognizer = recognizer
-    data.view = view
-    location = recognizer.locationInView_(view.objc_instance)
-    data.location = ui.Point(location.x, location.y)
-    data.state = recognizer.state()
-    data.number_of_touches = recognizer.numberOfTouches()
+class ObjCPlus:
+    """ docgen-ignore """
     
-    if (is_objc_type(recognizer, UIPanGestureRecognizer) or 
-    is_objc_type(recognizer, UIScreenEdgePanGestureRecognizer)):
-        trans = recognizer.translationInView_(ObjCInstance(view))
-        vel = recognizer.velocityInView_(ObjCInstance(view))
-        data.translation = ui.Point(trans.x, trans.y)
-        data.velocity = ui.Point(vel.x, vel.y)
-    elif is_objc_type(recognizer, UIPinchGestureRecognizer):
-        data.scale = recognizer.scale()
-        data.velocity = recognizer.velocity()
-    elif is_objc_type(recognizer, UIRotationGestureRecognizer):
-        data.rotation = recognizer.rotation()
-        data.velocity = recognizer.velocity()
+    def __new__(cls, *args, **kwargs):
+        objc_class = getattr(cls, '_objc_class', None)
+        if objc_class is None:
+            objc_class_name = cls.__name__ + '_ObjC'
+            objc_superclass = getattr(
+                cls, '_objc_superclass', NSObject)
+            objc_debug = getattr(cls, '_objc_debug', True)
+            
+            #'TempClass_'+str(uuid.uuid4())[-12:]
+            
+            objc_methods = []
+            objc_classmethods = []
+            for key in cls.__dict__:
+                value = getattr(cls, key)
+                if (inspect.isfunction(value) and 
+                    '_self' in inspect.signature(value).parameters
+                ):
+                    if getattr(value, '__self__', None) == cls:
+                        objc_classmethods.append(value)
+                    else:
+                        objc_methods.append(value)
+            if ObjCDelegate in cls.__mro__:
+                objc_protocols = cls.__name__
+            else:
+                objc_protocols = getattr(cls, '_objc_protocols', [])
+            if not type(objc_protocols) is list:
+                objc_protocols = [objc_protocols]
+            cls._objc_class = objc_class = create_objc_class(
+                objc_class_name,
+                superclass=objc_superclass,
+                methods=objc_methods,
+                classmethods=objc_classmethods,
+                protocols=objc_protocols,
+                debug=objc_debug
+            )
+        
+        instance = objc_class.alloc().init()
 
-    handler_func(data)
+        for key in dir(cls):
+            value = getattr(cls, key)
+            if inspect.isfunction(value):
+                if (not key.startswith('__') and 
+                not '_self' in inspect.signature(value).parameters):
+                    setattr(instance, key, types.MethodType(value, instance))
+                if key == '__init__':
+                    value(instance, *args, **kwargs)
+
+        return instance
+
+        
+class ObjCDelegate(ObjCPlus):
+    """ If you inherit from this class, the class name must match the delegate 
+    protocol name. (docgen-ignore) """
+            
+            
+def _is_objc_type(objc_instance, objc_class):
+    return objc_instance.isKindOfClass_(objc_class.ptr)
+
+
+class UIGestureRecognizerDelegate(ObjCDelegate):
+    """ docgen-ignore """
     
-def gestureRecognizer_shouldRecognizeSimultaneouslyWithGestureRecognizer_(
-        _self, _sel, _gr, _other_gr):
-    slf = ObjCInstance(_self)
-    other_gr = ObjCInstance(_other_gr)
-    return other_gr in slf.other_recognizers
+    def __init__(self, recognizer_class, view, handler_func):
+        self.view = view
+        self.handler_func = handler_func
+        self.other_recognizers = []
+        
+        view.touch_enabled = True
 
-GestureHandler = create_objc_class(
-    'GestureHandler',
-    superclass=NSObject,
-    methods=[
-        gestureAction,
-        gestureRecognizer_shouldRecognizeSimultaneouslyWithGestureRecognizer_
-    ],
-    protocols=['UIGestureRecognizerDelegate']
-)
-
-def _get_handler(recognizer_class, view, handler_func):
-    view.touch_enabled = True
-    handler = GestureHandler.new().autorelease()
-    retain_global(handler)
-
-    if handler_func == 'close':
-        recognizer = replace_close_gesture(view, recognizer_class)
-    else:
-        recognizer = recognizer_class.alloc().initWithTarget_action_(
-            handler, 'gestureAction').autorelease()
-        view.objc_instance.addGestureRecognizer_(recognizer)
-
-    handler.view = view
-    handler.recognizer = recognizer
-    handler.handler_func = handler_func
-    handler.other_recognizers = []
+        if handler_func == 'close':
+            self.recognizer = replace_close_gesture(view, recognizer_class)
+        else:
+            self.recognizer = recognizer_class.alloc().initWithTarget_action_(
+                self, 'gestureAction').autorelease()
+            view.objc_instance.addGestureRecognizer_(self.recognizer)
 
+        retain_global(self)
+    
+    def gestureAction(_self, _cmd):
+        self = ObjCInstance(_self)
+        view = self.view
+        recognizer = self.recognizer
+        handler_func = self.handler_func
+        data = Data()
+        data.recognizer = recognizer
+        data.view = view
+        location = recognizer.locationInView_(view.objc_instance)
+        data.location = ui.Point(location.x, location.y)
+        data.state = recognizer.state()
+        data.number_of_touches = recognizer.numberOfTouches()
+        
+        if (is_objc_type(recognizer, UIPanGestureRecognizer) or 
+        _is_objc_type(recognizer, UIScreenEdgePanGestureRecognizer)):
+            trans = recognizer.translationInView_(ObjCInstance(view))
+            vel = recognizer.velocityInView_(ObjCInstance(view))
+            data.translation = ui.Point(trans.x, trans.y)
+            data.velocity = ui.Point(vel.x, vel.y)
+        elif _is_objc_type(recognizer, UIPinchGestureRecognizer):
+            data.scale = recognizer.scale()
+            data.velocity = recognizer.velocity()
+        elif _is_objc_type(recognizer, UIRotationGestureRecognizer):
+            data.rotation = recognizer.rotation()
+            data.velocity = recognizer.velocity()
+    
+        handler_func(data)
+        
+    def gestureRecognizer_shouldRecognizeSimultaneouslyWithGestureRecognizer_(
+            _self, _sel, _gr, _other_gr):
+        self = ObjCInstance(_self)
+        other_gr = ObjCInstance(_other_gr)
+        return other_gr in self.other_recognizers
+        
     @on_main_thread
     def before(self, other):
         other.recognizer.requireGestureRecognizerToFail_(
@@ -307,12 +406,9 @@ def _get_handler(recognizer_class, view, handler_func):
     def together_with(self, other):
         self.other_recognizers.append(other.recognizer)
         self.recognizer.delegate = self
-            
-    setattr(handler, 'before', types.MethodType(before, handler))
-    setattr(handler, 'after', types.MethodType(after, handler))
-    setattr(handler, 'together_with', types.MethodType(together_with, handler))
 
-    return handler
+        
+#docgen: Gestures
 
 @on_main_thread
 def tap(view, action, 
@@ -326,7 +422,7 @@ def tap(view, action,
     * `number_of_touches_required` - Set if more than one finger is
       required for the gesture to be recognized.
     """
-    handler = _get_handler(UITapGestureRecognizer, view, action)
+    handler = UIGestureRecognizerDelegate(UITapGestureRecognizer, view, action)
 
     recognizer = handler.recognizer
     if number_of_taps_required:
@@ -367,7 +463,7 @@ def long_press(view, action,
     * `allowable_movement` - Set to change the default 10 point maximum
     distance allowed for the gesture to be recognized.
     """
-    handler = _get_handler(UILongPressGestureRecognizer, view, action)
+    handler = UIGestureRecognizerDelegate(UILongPressGestureRecognizer, view, action)
 
     recognizer = handler.recognizer
     if number_of_taps_required:
@@ -401,7 +497,7 @@ def pan(view, action,
     * `velocity` - Current velocity of the pan gesture as points per
       second (a `ui.Point` with `x` and `y` attributes).
     """
-    handler = _get_handler(UIPanGestureRecognizer, view, action)
+    handler = UIGestureRecognizerDelegate(UIPanGestureRecognizer, view, action)
 
     recognizer = handler.recognizer
     if minimum_number_of_touches:
@@ -425,7 +521,7 @@ def edge_pan(view, action, edges):
     Handler `action` receives the same gesture-specific attributes in
     the `data` argument as pan gestures, see `pan`.
     """
-    handler = _get_handler(UIScreenEdgePanGestureRecognizer, view, action)
+    handler = UIGestureRecognizerDelegate(UIScreenEdgePanGestureRecognizer, view, action)
 
     handler.recognizer.edges = edges
 
@@ -444,7 +540,7 @@ def pinch(view, action):
     * `velocity` - Current velocity of the pinch gesture as scale
       per second.
     """
-    handler = _get_handler(UIPinchGestureRecognizer, view, action)
+    handler = UIGestureRecognizerDelegate(UIPinchGestureRecognizer, view, action)
 
     return handler
 
@@ -461,7 +557,7 @@ def rotation(view, action):
     * `velocity` - Current velocity of the rotation gesture as radians
       per second.
     """
-    handler = _get_handler(UIRotationGestureRecognizer, view, action)
+    handler = UIGestureRecognizerDelegate(UIRotationGestureRecognizer, view, action)
 
     return handler
 
@@ -491,7 +587,7 @@ def swipe(view, action,
     multiple recognizers if you need to differentiate between the
     directions.
     """
-    handler = _get_handler(UISwipeGestureRecognizer, view, action)
+    handler = UIGestureRecognizerDelegate(UISwipeGestureRecognizer, view, action)
 
     recognizer = handler.recognizer
     if direction:
@@ -510,6 +606,9 @@ def swipe(view, action,
 
     return handler
 
+
+#docgen: Gesture management
+
 @on_main_thread
 def disable(handler):
     """ Disable a recognizer temporarily. """
@@ -559,6 +658,209 @@ def replace_close_gesture(view, recognizer_class):
     return recognizer
 
 
+# Drag and drop delegates
+
+drag_and_drop_prefix = 'py_object_'
+
+def _to_pyobject(item):
+    item = ObjCInstance(item)
+    try:
+        data = item.localObject()
+        if data is None: return None
+        if not str(data).startswith(drag_and_drop_prefix):
+            return None
+        address_str = str(data)[len(drag_and_drop_prefix):]
+        address = int(address_str)
+        result = ctypes.cast(address, ctypes.py_object).value
+        return result
+    except Exception as e:
+        return None
+
+
+class UIDragInteractionDelegate(ObjCDelegate):
+    """ docgen-ignore """
+    
+    def __init__(self, view, data, allow_others):
+        if not callable(data):
+            data = functools.partial(lambda d, sender: d, data)
+        self.data = { 'payload_func': data }
+        self.view = view
+        view.touch_enabled = True
+        draginteraction = UIDragInteraction.alloc().initWithDelegate_(self)
+        draginteraction.setEnabled(True)
+        draginteraction.setAllowsSimultaneousRecognitionDuringLift_(allow_others)
+        view.objc_instance.addInteraction(draginteraction)
+            
+        retain_global(self)
+    
+    def dragInteraction_itemsForBeginningSession_(_self, _cmd,
+    _interaction, _session):
+        self = ObjCInstance(_self)
+        session = ObjCInstance(_session)
+        payload = self.data['payload_func'](self.view)
+        # Retain reference to potentially ephemeral data
+        
+        self.content_actual = {
+            'payload': payload,
+            'sender': self.view
+        }
+        
+        external_payload = ''
+        
+        if type(payload) is str:
+            external_payload = payload
+        elif type(payload) in [ui.Image]:
+            external_payload = ObjCInstance(payload)
+        provider = NSItemProvider.alloc().initWithObject(external_payload)
+        item = UIDragItem.alloc().initWithItemProvider(provider)
+        item.setLocalObject_(
+            str(drag_and_drop_prefix) +  
+            str(id(self.content_actual)))
+        object_array = NSArray.arrayWithObject(item)
+        return object_array.ptr
+   
+
+class UIDropInteractionDelegate(ObjCDelegate):
+    """ docgen-ignore """
+    
+    def __init__(self, view, handler_func, accept=None):
+        self.accept_type = None
+        if type(accept) is type:
+            if accept is str:
+                self.accept_type = NSString
+            elif accept is ui.Image:
+                self.accept_type = UIImage
+            accept = functools.partial(
+                lambda dtype, d, s, r: type(d) is dtype, accept)
+        self.functions = {
+            'handler': handler_func,
+            'accept': accept
+        }
+        self.view = view
+        view.touch_enabled = True
+        
+        dropinteraction = UIDropInteraction.alloc().initWithDelegate_(self)
+        view.objc_instance.addInteraction(dropinteraction)
+        retain_global(self)
+        
+    def dropInteraction_canHandleSession_(_self, _cmd, _interaction, _session):
+        return True
+        
+    def dropInteraction_sessionDidUpdate_(_self, _cmd, _interaction, _session):
+        self = ObjCInstance(_self)
+        session = ObjCInstance(_session)
+        proposal = 2 # UIDropOperationCopy
+        accept_func = self.functions['accept']
+
+        if session.localDragSession():
+            if accept_func is not None:
+                for item in session.items():
+                    data = _to_pyobject(item)
+                    payload = data['payload']
+                    sender = data['sender']
+                    if not accept_func(payload, sender, self.view):
+                        proposal = 1 # UIDropOperationForbidden
+        else:
+            if (self.accept_type is None or
+            not session.canLoadObjectsOfClass(self.accept_type)):
+                    proposal = 1 # UIDropOperationForbidden
+
+        return UIDropProposal.alloc().initWithDropOperation(proposal).ptr
+        
+    def dropInteraction_performDrop_(_self, _cmd, _interaction, _session):
+        self = ObjCInstance(_self)
+        session = ObjCInstance(_session)
+        handler = self.functions['handler']
+        
+        if session.localDragSession():
+            for item in session.items():
+                data = _to_pyobject(item)
+                payload = data['payload']
+                sender = data['sender']
+                handler(payload, sender, self.view)
+        else:
+            if self.accept_type is not None:
+                
+                def completion_handler(_cmd, _object, _error):
+                    obj = ObjCInstance(_object)
+                    payload = None
+                    if _is_objc_type(obj, NSString):
+                        payload = str(obj)
+                    elif _is_objc_type(obj, UIImage):
+                        payload = ui.Image.from_data(uiimage_to_png(obj))
+                    handler(payload, None, self.view)
+                handler_block = ObjCBlock(
+                    completion_handler, restype=None,
+                    argtypes=[c_void_p, c_void_p, c_void_p])
+                retain_global(handler_block)
+                
+                for item in session.items():
+                    provider = item.itemProvider()
+                    provider.loadObjectOfClass_completionHandler_(
+                        self.accept_type, handler_block)
+                    break
+
+#docgen: Drag and drop                                
+        
+@on_main_thread
+def drag(view, payload, allow_others=False):
+    """ Sets the `view` to be the sender in a drag and drop operation. Dragging
+    starts with a long press.
+    
+    For within-app drag and drop, `payload` can be anything, and it is passed
+    by reference.
+    
+    If the `payload` is a text string or a `ui.Image`, it can be dragged
+    (copied) to another app (on iPad).
+    There is also built-in support for dropping text to any `ui.TextField` or
+    `ui.TextView`. 
+    
+    If `payload` is a function, it is called at the time when the drag starts.
+    The function receives one argument, the sending `view`, and must return the
+    data to be dragged.
+
+    Additional parameters:
+
+    * `allow_others` - Set to True if other gestures attached to the view
+    should be prioritized over the dragging.
+    """
+    
+    UIDragInteractionDelegate(view, payload, allow_others)
+    
+@on_main_thread
+def drop(view, action, accept=None):
+    """ Sets the `view` as a drop target, calling the `action` function with
+    dropped data.
+    
+    Additional parameters:
+
+    * `accept` - Control which data will be accepted for dropping. Simplest
+    option is to provide an accepted Python type like `dict` or `ui.Label`.
+    
+      For cross-app drops, only two types are currently supported: `str` for
+      plain text, and `ui.Image` for images.
+      
+      For in-app drops, the `accept` argument can also be a function that will
+      be called when a drag enters the view. Function gets same parameters
+      as the main handler, and should return False if the view should not accept
+      the drop.
+    
+    `action` function has to have this signature:
+        
+        def handle_drop(data, sender, receiver):
+            ...
+            
+    Arguments of the `action` function are:
+            
+    * `data` - The dragged data.
+    * `sender` - Source view of the drag and drop. This is `None` for drags
+    between apps.
+    * `receiver` - Same as `view`.
+    """
+    
+    UIDropInteractionDelegate(view, action, accept)
+
+
 if __name__ == '__main__':
 
     import math, random, console
@@ -648,7 +950,7 @@ if __name__ == '__main__':
 
     label_count = -1
 
-    def create_label(title):
+    def create_label(title, instance=None):
         global label_count
         label_count += 1
         label_w = 175
@@ -661,19 +963,22 @@ if __name__ == '__main__':
         line = math.floor(label_count / labels_per_line)
         column = label_count - line * labels_per_line
 
-        l = ui.Label(
-            text=title,
-            background_color='grey',
-            text_color='white',
-            alignment=ui.ALIGN_CENTER,
-            number_of_lines=0,
-            frame=(
-                left_margin + column * label_w_with_gap,
-                gap + line * label_h_with_gap,
-                label_w, label_h
-            ))
-        v.add_subview(l)
-        return l
+        if instance is None:
+            instance = ui.Label(
+                text=title,
+                background_color='grey',
+                text_color='white',
+                alignment=ui.ALIGN_CENTER,
+                number_of_lines=0
+            )
+        
+        instance.frame = (
+            left_margin + column * label_w_with_gap,
+            gap + line * label_h_with_gap,
+            label_w, label_h
+        )
+        v.add_subview(instance)
+        return instance
 
 
     tap_l = create_label('Tap')
@@ -709,3 +1014,33 @@ if __name__ == '__main__':
         maximum_number_of_touches=2)
     pinch_r = pinch(pan_and_pinch_l, pan_and_pinch_handler)
     pan_r.together_with(pinch_r)
+ 
+    drag_dict_l = create_label('Drag dict')
+    drag_image_l = create_label('Drag image')
+    drop_dict_l = create_label('Drop dict')
+    iv = ui.ImageView(
+        image=ui.Image('iow:image_32'),
+        background_color='grey',
+        content_mode=ui.CONTENT_CENTER,
+    )
+    iv.objc_instance.setClipsToBounds_(True)
+    drop_image_l = create_label('Image drop', iv)
+ 
+    drag(drag_dict_l, {'message': 'Success'})
+    
+    drag(drag_image_l, ui.Image('iow:ios7_checkmark_32'))
+    
+    def dict_dropped(data, sender, receiver):
+        receiver.text = f"Drop\n{data['message']}"
+        ui.delay(lambda: setattr(receiver, 'text', 'Drop'), 2.0)
+
+    drop(drop_dict_l, dict_dropped, accept=dict)
+    
+    def image_dropped(data, sender, receiver):
+        receiver.image = data
+        ui.delay(lambda:
+            setattr(receiver, 'image', ui.Image('iow:image_24')),
+            2.0
+        )
+        
+    drop(drop_image_l, image_dropped, accept=ui.Image)
